from .closed_QA import *
from .coreference import *
from .entity_typing import *
from .generation import *
from .nli import *
from .paraphrase import *
from .reading_comprehensation import *
from .relation import *
from .sentiment import *
from .summarization import *
from .topic_classification import *